const claudia = require('claudia');
const path = require('path');
const AWS = require('aws-sdk');
const { util } = require('../util');
const { ConsoleLogger } = require('./lib');
const fs = require('fs-extra');

const CLOUD_RED_DEPLOYMENTS_BUCKET = 'cloud-red-deployments'; // only used for debugging purposes
const CLAUDIA_CONFIG_FILENAME = 'cloud-red-config.json';

/**
 * @property {string=} userDir - directory where the user is running the project on
 * @property {string=} lambdaName - (optional) Lambda name (default. based on the property name of package.json)
 * @property {string=} lambdaHandler - (optional) Lambda handler. Default: `index.handler`
 * @property {string=} version - (optional) A version alias to automatically assign to the new function.
 * @property {string} profile - AWS profile to be used
 * @property {string} region - AWS region to be used. Default: `ap-southeast-2`
 * @property {string=} timeout - (optional) Lambda Timeout in secs. Default: `30`
 * @property {string=} memory - (optional) The amount of memory, in MB, your Lambda function is given. Default: `256`
 * @property {string=} runtime - (optional) Node.js runtime to use. For supported values, see http://docs.aws.amazon.com/lambda/latest/dg/API_CreateFunction.html. Default: `nodejs10.x`
 * @property {string=} env - (optional) comma-separated list of VAR=VALUE environment variables to set. For example: S3BUCKET=testbucket,SNSQUEUE=testqueue
 */
class AWSProvider {
  constructor(
    userDir,
    lambdaName,
    lambdaHandler,
    version,
    profile,
    region,
    timeout,
    memory,
    runtime,
    env
  ) {
    this.userDir = userDir;
    this.lambdaName = lambdaName;
    this.lambdaHandler = lambdaHandler || 'handler';
    this.version = version;
    this.profile = profile;
    this.region = region;
    this.timeout = timeout || 30;
    this.memory = memory || 256;
    this.runtime = runtime || 'nodejs10.x';
    this.env = env;
  }

  /**
   * Returns true if the newFlow contains at least one apigw-eventhandler
   * @param {CloudRed Flow} newFlow
   */
  isDeployProxyAPIRequested(newFlow) {
    const apigwHandlers = newFlow.filter(
      kind => kind.type === 'apigw-event-handler'
    );
    return apigwHandlers.length > 0;
  }
  // create a lambda deployment
  async createDeployment(newFlow) {
    // clean up previous deployment if required
    const status = await this.destroyDeployment();
    if (!status) {
      process.exit(1);
    }

    // bundle the opts for claudia
    let claudiaOpts = {
      region: this.region,
      name: this.lambdaName || this.defaultLambdaName(),
      source: this.userDir,
      handler: this.lambdaHandler,
      timeout: this.timeout,
      memory: this.memory,
      config: path.join(this.userDir, CLAUDIA_CONFIG_FILENAME)
    };
    if (this.version) {
      claudiaOpts['version'] = this.version;
    }

    claudiaOpts['use-local-dependencies'] = true;

    // Uncomment below line only for debugging purposes
    // claudiaOpts['use-s3-bucket'] = CLOUD_RED_DEPLOYMENTS_BUCKET;

    if (this.env) {
      claudiaOpts['set-env'] = this.env;
    }

    if (this.isDeployProxyAPIRequested(newFlow)) {
      claudiaOpts['deploy-proxy-api'] = true;
    }

    // inject the profile to be used to aws
    AWS.config.credentials = new AWS.SharedIniFileCredentials({
      profile: this.profile
    });

    try {
      let output = await claudia.create(claudiaOpts, new ConsoleLogger());
      console.log(JSON.stringify(output, null, 2));
      return true;
    } catch (err) {
      console.log(err);
      process.exit(1);
    }
  }

  /**
   * return {boolean} status - `true` if destroyed or not required; `false` otherwise
   */
  async destroyDeployment() {
    let success = false;
    const prevDeploymentCfg = path.join(this.userDir, CLAUDIA_CONFIG_FILENAME);
    if (fs.existsSync(prevDeploymentCfg)) {
      console.log(
        `Cleaning up resources defined in file: ${CLAUDIA_CONFIG_FILENAME}...`
      );

      // Load credentials
      AWS.config.credentials = new AWS.SharedIniFileCredentials({
        profile: this.profile
      });

      let claudiaOpts = {
        source: this.userDir,
        config: prevDeploymentCfg
      };
      try {
        await claudia.destroy(claudiaOpts);
        success = true;
      } catch (err) {
        console.log(err);
        success = false;
      }
    } else {
      console.log(
        `No "${CLAUDIA_CONFIG_FILENAME}" file found. No resources to be clean up.`
      );
      success = true;
    }

    return success;
  }

  async fetchLambdaLogs(lambdaName) {
    //TODO: read from cloudwatch and send to debug logs ...
    throw new Error('Not implemented yet');
  }

  defaultLambdaName() {
    if (this.lambdaName) {
      return this.lambdaName;
    }
    // generate a default lambda name based on the `name` property of the project's package.json
    const pkg = util.readJSONSync(path.join(this.userDir, 'package.json'));
    return this.lambdaNameSanitize(pkg.name);
  }
  /**
   * clean up the lambda name to make it aws compatible
   * @param {string} str : Original name for the lambda.
   */
  lambdaNameSanitize(original) {
    'use strict';
    return (
      original &&
      original
        .trim()
        .replace(/^@/, '')
        .replace(/[^a-zA-Z0-9-_]/g, '_')
        .substr(0, 140)
    );
  }
}

module.exports.AWSProvider = AWSProvider;
